/* index sequential files, producing .idx and .sel files */
/* Bruce Tanner - Cerritos College */

/* 1.0  05/10/93 Original attempt */
/* 2.0  06/20/93 Create indexed files directly, add keyword count field */
/* 2.1  07/08/93 Change the file name for NOISE_WORDS */
/* 2.2  07/08/93 Move the range end (end_pos) to before the terminator */
/* 2.2jlw 07/14/93 - JLW added length spec to dash, added additional topic
                     divider keywords */
/* 2.3  07/19/93 Set multi-buffer, multi-block counts, read-ahead, write-behind
                and deferred write; noticeably increased performance */
/* 2.4  07/26/93 Removed index name, added CLI$ interface, added /topic */
/* 2.4jlw 07/27/93 fixed version retention, which was broken */
/* 2.5  07/27/93 Selector strings forced to lowercase; use a good copy */
/* 2.6  07/29/93 revamp /topic syntax to include text, size, exclude */
/* 2.7  07/30/93 make size=n pad as well as truncate field width */
/* 2.8  08/03/93 take wildcard input file names, add /output, /version */
/* 2.9  08/05/93 JLW changed filename sizes from 80 to 256 characters */
/* 2.10 08/05/93 add check for max number of topics, reformat code */
/* 2.11 08/24/93 JLW added specific statuses for exit errors */

#include ssdef
#include stdio
#include string
#include ctype
#include rms
#include descrip
#include climsgdef

#define CHUNK 50     /* increment to expand table of words */
#define DESC_SIZE 70 /* maximum size of a topic description */
#define SELECTOR_SIZE 100 /* maximum size of a selector (minus description) */
#define SELECTOR_INDEX 4 /* number of digits for selector number */
#define COUNT_SIZE SELECTOR_INDEX /* number of digits for keyword count field */
#define MAX_TOPIC 20 /* maximum number of topics to list */

typedef struct {
    char *text;
    char *found;
    int  size;
    int  exclude;
} topic_str;

int words_index, words_size;
char **words;
int noise_index, noise_size;
char **noise;
int max_word;
char *keyword, *idx_record;
char selector_number[SELECTOR_INDEX + 1];
char sel_record[SELECTOR_INDEX + DESC_SIZE + SELECTOR_SIZE + 1];

void build_words(char *);
void expand_table(char ***, int *);
void write_words(FILE *, struct RAB *, struct RAB*, int *, int *, int *,
                 char *, topic_str *);
void load_noise();
int is_noise(char *, int);
struct dsc$descriptor_s *descr (char *string);
void parse_topic(char *, topic_str *);
void *my_realloc(void *, int);
void index_commands();
int lib$get_foreign(), lib$get_input();


main(int argc, char *argv[])
{

    FILE  *src;
    char  *ptr, desc[DESC_SIZE + 1], src_line[256], cli_input[256];
    char  value[20], file_arg[256], file_spec[256], out_name[256];
    char  orig_line[256], spaces[DESC_SIZE + 1];
    int   start_pos, end_pos, db_index, status, index, context;
    enum  {para, dash, hex, equal, line, whole} type = para;
    int   dash_len;
    int   hex_value;
    short leng;
    char  *dashes = NULL;
    struct FAB idxfab, selfab;
    struct RAB idxrab, selrab;
    struct XABKEY idxxab, selxab;
    topic_str topics[MAX_TOPIC];
    $DESCRIPTOR(input_dsc, cli_input);
    $DESCRIPTOR(file_dsc, file_arg);
    $DESCRIPTOR(file_spec_dsc, file_spec);
    $DESCRIPTOR(out_dsc, out_name);
    $DESCRIPTOR(value_dsc, value);

    status = lib$get_foreign(&input_dsc, 0, &leng, 0);

    strncpy(cli_input+6, cli_input, leng);
    strncpy(cli_input, "index ", 6);

    status = cli$dcl_parse(&input_dsc, index_commands, lib$get_input);

    if (status != CLI$_NORMAL)  /* error in parse, exit */
        exit(7);

    if ((cli$present(descr("file")) & 1) == 0) {
        printf("Usage: index document\n");
        printf(" /TOPIC=(text,size,exclude)  capture lines that start with keyword\n");
        printf(" /WORD_LENGTH=n  maximum size of index key (default 20)\n");
        printf(" /PARAGRAPH      text separated by blank lines\n");
        printf(" /FF             text separated by form feeds\n");
        printf(" /DASH=n         text separated n dashes (default 3)\n");
        printf(" /EQUAL=n        text separated n equals (default 80)\n");
        printf(" /CHARACTER=n    text separated by control character 'n'\n");
        printf(" /LINE           each line is separate text entry\n");
        printf(" /WHOLE          whole file is one text entry\n");
        printf(" /OUTPUT=file    override name of index/selection files\n");
        printf(" /[NO]VERSION    keep or discard document version in selection\n");
        exit(1);
    }

    if (cli$present(descr("paragraph")) & 1)
        type = para;
    if (cli$present(descr("FF")) & 1) {
        type = hex;                      /* /FF same as /character=12 */
        hex_value = '\f';
    }
    if (cli$present(descr("character")) & 1) {
        status = cli$get_value(descr("character"), &value_dsc, &leng);
        type = hex;
        value[leng] = '\0';
        hex_value = atoi(value);
    }
    if (cli$present(descr("whole")) & 1)
        type = whole;
    if (cli$present(descr("line")) & 1)
        type = line;
    if (cli$present(descr("dash")) & 1) {
        status = cli$get_value(descr("dash"), &value_dsc, &leng);
        type = dash;
        value[leng] = '\0';
        dash_len = atoi(value);
    }
    if (cli$present(descr("equal")) & 1) {
        status = cli$get_value(descr("equal"), &value_dsc, &leng);
        type = equal;
        value[leng] = '\0';
        dash_len = atoi(value);
    }
    if ((status = cli$present(descr("word_length"))) & 1) {
        status = cli$get_value(descr("word_length"), &value_dsc, &leng);
        value[leng] = '\0';
        max_word = atoi(value);
    }
    for (index = 0; index < MAX_TOPIC; index++) {
        topics[index].text = NULL;
        topics[index].found = NULL;
        topics[index].size = 0;
        topics[index].exclude = 0;
    }
    parse_topic(cli_input, topics); /* parse the command line and fill topics */

    status = cli$get_value(descr("file"), &file_dsc, &leng);  /* get source */

    status = lib$find_file(&file_dsc, &file_spec_dsc, &context, 0, 0, 0, 0);
    ptr = strchr(file_spec, ' ');
    if (ptr) *ptr = '\0';            /* chop off trailing spaces */
    strcpy(out_name, file_spec);    /* make copy for output spec */

    if (cli$present(descr("output")) & 1) { /* if /output, overwrite out_name */
        status = cli$get_value(descr("output"), &out_dsc, &leng);
        out_name[leng] = '\0';
    }

    words_size = words_index = 0;    /* no words yet */
    words = NULL;
    noise_size = noise_index = 0;    /* no noise yet */
    noise = NULL;
    load_noise();  /* build a list of words to ignore */
    db_index = 0;

    dashes = (char *)malloc(dash_len+1);
    memset((void *)dashes, (type==dash)?'-':'=', dash_len);
    dashes[dash_len] = '\0';

    memset((void *) spaces, ' ', DESC_SIZE); /* make spaces for padding topic */
    spaces[DESC_SIZE] = '\0';

    keyword = (char *) calloc(max_word + SELECTOR_INDEX + 1, sizeof(char));
    idx_record = (char *) calloc(max_word + SELECTOR_INDEX + COUNT_SIZE + 1,
                                 sizeof(char));

    ptr = strrchr(out_name, '.');  /* just get file name */
    if (ptr) *ptr = '\0';
    strcat(out_name, ".idx");

    idxfab = cc$rms_fab;
    idxfab.fab$l_alq = 50;
    idxfab.fab$b_bks = 3;
    idxfab.fab$w_deq = 10;
    idxfab.fab$b_fac = FAB$M_GET | FAB$M_PUT | FAB$M_UPD;
    idxfab.fab$l_fna = out_name;
    idxfab.fab$b_fns = strlen(out_name);
    idxfab.fab$l_fop = FAB$M_CBT | FAB$M_DFW;
    idxfab.fab$w_mrs = max_word + SELECTOR_INDEX + COUNT_SIZE;
    idxfab.fab$b_org = FAB$C_IDX;
    idxfab.fab$b_rat = FAB$M_CR;
    idxfab.fab$b_rfm = FAB$C_FIX;
    idxfab.fab$b_shr = FAB$M_NIL;
    idxfab.fab$l_xab = &idxxab;

    idxrab = cc$rms_rab;
    idxrab.rab$l_fab = &idxfab;
    idxrab.rab$b_krf = 0;
    idxrab.rab$l_kbf = keyword;
    idxrab.rab$b_ksz = max_word + SELECTOR_INDEX;
    idxrab.rab$b_rac = RAB$C_KEY;
    idxrab.rab$l_rbf = idx_record;
    idxrab.rab$w_rsz = max_word + SELECTOR_INDEX + COUNT_SIZE;
    idxrab.rab$l_ubf = idx_record;
    idxrab.rab$w_usz = max_word + SELECTOR_INDEX + COUNT_SIZE;
    idxrab.rab$b_mbf = 20;
    idxrab.rab$l_rop = RAB$M_RAH | RAB$M_WBH;

    idxxab = cc$rms_xabkey;
    idxxab.xab$b_dtp = XAB$C_STG;
    idxxab.xab$b_flg = XAB$M_DAT_NCMPR | XAB$M_IDX_NCMPR;
    idxxab.xab$w_pos0 = 0;
    idxxab.xab$b_siz0 = max_word + SELECTOR_INDEX;
    idxxab.xab$b_ref = 0;

    if (((status = sys$create(&idxfab)) & 1) != SS$_NORMAL)
        lib$stop(status);
    if (((status = sys$connect(&idxrab)) & 1) != SS$_NORMAL)
        lib$stop(status);

    ptr = strrchr(out_name, '.');  /* just get file name */
    if (ptr) *ptr = '\0';
    strcat(out_name, ".sel");

    selfab = cc$rms_fab;
    selfab.fab$l_alq = 10;
    selfab.fab$b_bks = 3;
    selfab.fab$w_deq = 5;
    selfab.fab$b_fac = FAB$M_PUT;
    selfab.fab$l_fna = out_name;
    selfab.fab$b_fns = strlen(out_name);
    selfab.fab$l_fop = FAB$M_CBT | FAB$M_DFW;
    selfab.fab$w_mrs = SELECTOR_INDEX + DESC_SIZE + SELECTOR_SIZE;
    selfab.fab$b_org = FAB$C_IDX;
    selfab.fab$b_rat = FAB$M_CR;
    selfab.fab$b_rfm = FAB$C_VAR;
    selfab.fab$b_shr = FAB$M_NIL;
    selfab.fab$l_xab = &selxab;

    selrab = cc$rms_rab;
    selrab.rab$l_fab = &selfab;
    selrab.rab$b_krf = 0;
    selrab.rab$l_kbf = selector_number;
    selrab.rab$b_ksz = SELECTOR_INDEX;
    selrab.rab$b_rac = RAB$C_KEY;
    selrab.rab$l_rbf = sel_record;
    selrab.rab$b_mbf = 20;
    selrab.rab$l_rop = RAB$M_RAH | RAB$M_WBH;

    selxab = cc$rms_xabkey;
    selxab.xab$b_dtp = XAB$C_STG;
    selxab.xab$b_flg = XAB$M_DAT_NCMPR | XAB$M_IDX_NCMPR;
    selxab.xab$w_pos0 = 0;
    selxab.xab$b_siz0 = SELECTOR_INDEX;
    selxab.xab$b_ref = 0;

    if (((status = sys$create(&selfab)) & 1) != SS$_NORMAL)
        lib$stop(status);
    if (((status = sys$connect(&selrab)) & 1) != SS$_NORMAL)
        lib$stop(status);

    for (;;) {  /* process all files in input spec, first one already found */

        if ((src = fopen(file_spec, "r", "mbc=50", "mbf=20")) == NULL) { 
            printf("Can't read input file %s\n", file_spec);
            exit(3);
        }
        printf("Building index for %s\n", file_spec);

        start_pos = ftell(src);          /* init start position */
        strncpy(desc, "", DESC_SIZE + 1);
        
        while (fgets(src_line, sizeof(src_line), src)) {
            if ((src_line[0] == hex_value) && (type == hex)) {
                write_words(src, &selrab, &idxrab, &db_index, &start_pos,
                            &end_pos, desc, topics);
                continue;
            }
            ptr = strchr(src_line, '\n');
            if (ptr) *ptr-- = '\0';              /* remove newline */
            for (ptr = src_line; *ptr; ptr++)
                if (iscntrl(*ptr))  *ptr = ' ';  /* convert tabs to spaces */
            for (ptr = src_line; *ptr; ptr++)
                if (*ptr > ' ') break;  /* find first non-blank char */
            strcpy(src_line, ptr);      /* remove leading blanks */
            while ((strlen(src_line) > 0) &&
                   (src_line[strlen(src_line)-1] == ' '))
                src_line[strlen(src_line)-1] = '\0';/* remove trailing blanks */
            strcpy(orig_line, src_line);    /* copy before forcing lower case */
            for (ptr = src_line; *ptr; ptr++)
                *ptr = _tolower(*ptr);           /* force lowercase */
        
            if ((strncmp(src_line, dashes, dash_len) == 0) &&
                ((type == equal) ||(type == dash))) {
                write_words(src, &selrab, &idxrab, &db_index, &start_pos,
                            &end_pos, desc, topics);
                continue;
            }
            if ((strlen(src_line) == 0) && (type == para)) {
                write_words(src, &selrab, &idxrab, &db_index, &start_pos,
                            &end_pos, desc, topics);
                continue;
            }
            if (strlen(desc) == 0)       /* save the first line by default */ 
                strncpy(desc, orig_line, DESC_SIZE);
            for (index = 0; topics[index].text; index++) /* apply topic rules */
                if (strncmp(src_line, topics[index].text,
                            strlen(topics[index].text)) == 0) {
                    if (topics[index].exclude)
                        strcpy(orig_line, orig_line + strlen(topics[index].text));
                    topics[index].found = (char *) my_realloc((char *) topics[index].found,
                                         (topics[index].size ? topics[index].size : strlen(orig_line))
                                         + 1);
                    if (topics[index].size > 0) {
                        strncpy(topics[index].found, orig_line,
                                topics[index].size);
                        topics[index].found[topics[index].size] = '\0';
                        strncat(topics[index].found, spaces,
                                topics[index].size - strlen(topics[index].found));
                    }
                    else
                        strcpy(topics[index].found, orig_line);
                break;     /* a line satisfies only one topic rule */
                }
            build_words(src_line);
            end_pos = ftell(src);  /* end_pos points before any terminator */
            if (type == line) {
                write_words(src, &selrab, &idxrab, &db_index, &start_pos,
                            &end_pos, desc, topics);
                continue;
            }
        }
        
        /* in case file doesn't end with a terminator */
        write_words(src, &selrab, &idxrab, &db_index, &start_pos, &end_pos,
                    desc, topics);
        fclose(src);
        status = lib$find_file(&file_dsc, &file_spec_dsc, &context, 0, 0, 0, 0);
        if ((status & 1) == 0) {
            lib$find_file_end(&context);
            break;
        }
        ptr = strchr(file_spec, ' ');
        if (ptr) *ptr = '\0';            /* chop off trailing spaces */
    }
    sys$close(&selfab);
    sys$close(&idxfab);
}


/* break line into words and save them in words[] */

void build_words(char *line)
{
    char *cp, *cp2;

    for (cp = line; *cp; cp++)      /* convert punctuation to spaces */
        if (ispunct(*cp))  *cp = ' ';

    strcat(line, " ");              /* line ends with a space */
    cp = line;
    while(cp2 = strchr(cp, ' ')) {  /* break at space boundary */
        *cp2 = '\0';
        if (strlen(cp) > max_word)
            printf("Truncating %d character word /%s/ to %d characters\n",
                   strlen(cp), cp, max_word);
        if ((strlen(cp) > 0) && (!is_noise(cp, noise_size))) {
            if (words_index == words_size)  /* table full */
                expand_table(&words, &words_size);
            strncpy(words[words_index++], cp, max_word);
        }
        cp = cp2 + 1;
    }
}


/* expand *table[] by CHUNK elements of max_word characters */

void expand_table(char ***table, int *size)
{
    int ind;

    *table = (char **) my_realloc((char **) *table, (*size + CHUNK) * sizeof(char *));
    for (ind = 0; ind < CHUNK; ind++)
        (*table)[*size + ind] = (char *) calloc(max_word + 1, sizeof(char));
    *size += CHUNK;
}


/* write out Gopher command, write out words */

void write_words(FILE *src, struct RAB *selptr, struct RAB *idxptr,
                 int *db_index, int *start_pos, int *end_pos, char *desc,
                 topic_str *topics)
{
    int ind, status;
    char filename[256], count[COUNT_SIZE + 1], *ptr, temp_desc[512] = "";

    if (words_index == 0)
        return;      /* no words to write */
    (*db_index)++;
    if ((*db_index) == 10000) {
        printf("You have reached 10,000 topics in this index\n");
        printf("You will have to index fewer files or rebuild everything ");
        printf("with SELECTOR_INDEX larger\n");
        exit(5);
    }
    fgetname(src, filename);
    if ((cli$present(descr("version")) & 1) == 0) {  /* if /noversion */
        ptr = strchr(filename, ';');  /* get rid of version number */
        if (ptr) *ptr = '\0';
    }
    for(ptr = filename; *ptr; ptr++)
        *ptr = _tolower(*ptr);  /* force filename lowercase */
    for (ind = 0; ind < MAX_TOPIC; ind++)
        if (topics[ind].found && (strlen(topics[ind].found) > 0)) {
            if (strlen(temp_desc) > 0)
                strcat(temp_desc, " ");
            strcat(temp_desc, topics[ind].found);
        }
    if (strlen(temp_desc) > 0)
        strncpy(desc, temp_desc, DESC_SIZE);

    /* write out the selector */
    sprintf(selector_number, "%0*d", SELECTOR_INDEX, *db_index);
    sprintf(sel_record, "%0*d0%s\tR%d-%d-%s",
            SELECTOR_INDEX, *db_index, desc, *start_pos, *end_pos, filename);
    selptr->rab$l_rbf = sel_record;
    selptr->rab$w_rsz = strlen(sel_record);
    if (((status = sys$put(selptr)) & 1) != SS$_NORMAL)
        lib$stop(status);

    /* write out the words */
    for (ind = 0; ind < words_index; ind++) {
        sprintf(keyword, "%-*s%0*d", max_word, words[ind],
                SELECTOR_INDEX, *db_index);
        sprintf(idx_record, "%-*s%0*d%0*d",
                max_word, words[ind],
                SELECTOR_INDEX, *db_index,
                COUNT_SIZE, 1);
        status = sys$put(idxptr);
        /* record exists, increment count */
        if (status == RMS$_DUP) {
            status = sys$get(idxptr);
            if ((status & 1) != SS$_NORMAL)
                lib$stop(status);
            strncpy(count, idx_record + max_word + SELECTOR_INDEX, COUNT_SIZE);
            count[COUNT_SIZE] = '\0';
            sprintf(idx_record, "%-*s%0*d%0*d",
                    max_word, words[ind],
                    SELECTOR_INDEX, *db_index,
                    COUNT_SIZE, atoi(count) + 1);
            status = sys$update(idxptr);
            if ((status & 1) != SS$_NORMAL)
                lib$stop(status);
        }
        if ((status & 1) != SS$_NORMAL)
            lib$stop(status);
        *words[ind] = '\0';
    }
    printf("%s\n", desc);
    strncpy(desc, "", DESC_SIZE + 1);
    *start_pos = ftell(src);          /* init start position */
    words_index = 0;
    for (ind = 0; ind < MAX_TOPIC; ind++)
        if (topics[ind].found)
            *topics[ind].found = '\0';
}


/* read in a file of noise words, one per line */

void load_noise()
{
    FILE *nf;
    char *ptr, line[40];

    if ((nf = fopen("_noise_words", "r", "dna = gopher_root:[000000].dat")) == NULL)
        return;

    while (fgets(line, sizeof(line), nf)) {
        ptr = strchr(line, '\n');
        if (ptr) *ptr-- = '\0';         /* remove newline */
        if (strlen(ptr) > 0) {
            if (noise_index == noise_size)  /* table full */
                expand_table(&noise, &noise_size);
            strcpy(noise[noise_index++], line);
        }
    }

    fclose(nf);
}


/* see if the word is noise */

int is_noise(char *word, int size)
{
    int ind;

    if (strlen(word) <= 2)   /* simple heuristic saves lots of IO/memory */
        return(TRUE);
    for(ind = 0; ind < size; ind++)
        if (strcmp(noise[ind], word) == 0)
            return (TRUE);
    return (FALSE);
}


/* descr() creates character descriptor and return the address
of the descriptor to the caller. */
# define N_DESCR 10
static struct dsc$descriptor_s str_desc[N_DESCR];
static int cur_descr = -1;

struct dsc$descriptor_s *descr(char *string)
{
    if(++cur_descr >= N_DESCR) cur_descr = 0;
    str_desc[cur_descr].dsc$w_length=(short)strlen(string);      
    str_desc[cur_descr].dsc$b_dtype=DSC$K_DTYPE_T;   
    str_desc[cur_descr].dsc$b_class=DSC$K_CLASS_S;  
    str_desc[cur_descr].dsc$a_pointer=string;     
    return (&str_desc[cur_descr]);
}


/* parse command line for /topic */
void parse_topic(char *line, topic_str *topics)
{
    char *ptr, *start;
    static int index = -1;

    for (ptr = line; *ptr; ptr++)
        *ptr = _tolower(*ptr);         /* force command line lowercase */
    ptr = line;                        /* point to start of line */
    for (;;) {                         /* search for /topic until end of line */
        if (index == MAX_TOPIC)
            return;                    /* exit if we can't hold any more */
        ptr = strchr(ptr, '/');        /* search for switch start */
        if (ptr == NULL)
            return;                    /* no more switches */
        while (isspace(*++ptr));       /* skip spaces */
        if (*ptr != 't')               /* topic is unique to one character */
            continue;                  /* not /topic, keep scanning */
        do
            ptr++;
        while ((*ptr != '=')           /* skip to the keyword/parameter */
               && (*ptr != ':'));      /* separator character */
        while (isspace(*++ptr));       /* skip spaces */
        if (*ptr == '(')               /* if start of list */
            while (isspace(*++ptr));   /* skip spaces */
        index++;                       /* next topics structure */
        for (;;) {                     /* parse all /topic list elements */
            switch (*ptr) {
                case 't':                            /* text */
                    do 
                        ptr++;
                    while ((*ptr != '=')         /* skip to the keyword/parameter */
                           && (*ptr != ':'));    /* separator character */
                    while (isspace(*++ptr));     /* skip spaces */
                    if (*ptr == '"') {           /* if quoted string */
                        start = ++ptr;           /* skip over quote */
                        while (*ptr != '"')      /* skip to ending quote */
                            ptr++;
                    }
                    else {                       /* else non-quoted string */
                        start = ptr;             /* start of string */
                        while (*ptr
                               && (*ptr != ' ')
                               && (*ptr != ',')
                               && (*ptr != '/')
                               && (*ptr != ')'))
                            ptr++;               /* skip to string terminator */
                    }
                    topics[index].text = (char *) calloc((ptr - start) + 1, sizeof(char));
                    strncpy(topics[index].text, start, ptr - start);
                    break;
                case 's':                        /* size */
                    do
                        ptr++;
                    while ((*ptr != '=')       /* skip to the keyword/parameter */
                           && (*ptr != ':'));  /* separator character */
                    while (isspace(*++ptr));   /* skip spaces */
                    topics[index].size =       /* get (hopefully) decimal number */
                        (atoi(ptr) < DESC_SIZE) ? atoi(ptr) : DESC_SIZE;
                    break;
                case 'e':                         /* exclude */
                    topics[index].exclude = TRUE; /* has no parameters */
                    break;
            }
            while (*ptr
                   && (*ptr != ' ')              /* skip to end of */
                   && (*ptr != ',')              /* keyword */
                   && (*ptr != '/')              /* switch */
                   && (*ptr != ')'))             /* or parameter */
                ptr++;
            while (*ptr &&
                   ((*ptr <= ' ') ||             /* skip spaces, junk */
                    (*ptr == ',')))              /* list seperators */
                ptr++;
            if (*ptr == '\0')
                return;                          /* end of the line */
            if ((*ptr == ')') || (*ptr == '/'))
                break;                           /* end of the list */
        }                                        /* scan for more list elements */
    }
}


void *my_realloc(void *mem, int size)
{
    if (mem == (void *) 0)
        return ((void *) malloc(size));
    else
        return((void *) realloc(mem, size));
}

