/* Edit a file's Access Control List, accessing *only* GOPHER Lookaside
    entries.  Use callable EDT as the editor				*/
/* J.Lance Wilkinson, Penn State University Library Computing Services	*/

/* usage: lookaside [filename]						*/

/* 1.0 07/19/93 Original attempt */
/* 1.1 07/21/93 Added config file, eliminated journalling */
/* 1.2 07/23/93 Added new tuple entries Head=, Foot=, RHead=, RFoot= */

#include "gsgopherobj.h"	    /* GS_* */
#include <ctype.h>		    /* isspace(), etc.	*/
#include <stdio.h>		    /* NULL, stdin, etc... */
#include <string.h>		    /* C RTL string functions */
#include <atrdef.h>		    /* VMS File Attributes definitions */
#include <descrip.h>		    /* string descriptor defs. */
#include <ssdef.h>		    /* error codes, etc. */
#include <rms.h>		    /* RMS NAM defs */
#include <acedef.h>		    /* ACE defs */
#include <acldef.h>		    /* ACL defs */

int DEBUG = 0;
int access_tuple;

struct bpv_arg	    /* Bound Procedure Value */
{
    int *routine$;  /* Pointer to routine */
    int env$;	    /* Pointer to environment */
} ;

struct item	    /* Itemlist entry */
{
    short   buflen;
    short   code;
    char    *bufadr;
    int	    unused;
} ;
static	char	*g$ace[GS_TUPLE_CNT];

#pragma builtins

#define ACL_BUFFER_SIZE ATR$S_READACL /* ACL buffer size
                                         (512 is an XQP limit) */
#define _zero(_buffer_size, _buffer) \
{int i;\
_MOVC5((unsigned short) 0, (char *) 0, (char) 0,\
 (unsigned short) _buffer_size, _buffer, (unsigned short *) &i);\
}

static	struct FAB acl_fab;	/* storage for a FAB */
static	struct FAB lks_fab;	/* storage for another FAB */
static	struct RAB lks_rab;	/* storage for a RAB */
static	struct XABPRO acl_xab;	/* storage for an XABPRO */
static	struct NAM acl_nam;	/* storage for a NAM */

static	char	*acl_ptr=NULL;
static	char	*acl_new=NULL;
static	char	lookaside[256];


void
ZapCRLF(inputline)
  char *inputline;
{
     char *cp;

     cp = strchr(inputline, '\r');    /* Zap CR-LF */
     if (cp != NULL)
	  *cp = '\0';
     else {
	  cp = strchr(inputline, '\n');
	  if (cp != NULL)
	       *cp = '\0';
     }
}

/*
        Using a DCL foreign command line;

        argv[0] is the image name
        argv[1] is the file to edit ACLs for.
	argv[2] is an optional DEBUG keyword.

*/

main(int argc, char *argv[])                /* Main entry point */
{
    int	    status;
    char    *cp;
    char    ConfigFile[256];
    struct  
	dsc$descriptor_s
	    input_file, command_file;
    int	    acl_fileio(int *code, int *stream,
		    struct dsc$descriptor *record, struct dsc$descriptor *rhb);
globalvalue edt$m_nocreate;
globalvalue edt$m_nocommand;
globalvalue edt$m_nooutput;
globalvalue edt$m_nojournal;
globalvalue edt$m_command;
globalvalue edt$m_recover;
globalvalue edt$_inpfilnex;
    int	    options;
    FILE    *gdcfile;
    char    inputline[256];
    char    *token;
    char    *restofline;
static 
    struct 
	 bpv_arg 
	    fileio_bpv = {acl_fileio, 0};

    for (status=0; status<GS_TUPLE_CNT; g$ace[status++] = NULL);
    access_tuple = LinkTuple(GS_ACCS,NULL);

    if (argc < 2) {                  /* Verify that a file name was specified */
	argv[1] = (char *)malloc(256);
	printf("\nEnter File Specification: "),scanf("%s",argv[1]);
	fflush(stdin);
	}
    lookaside[0] = '\0';
    strcpy(ConfigFile, "GOPHER_CONFIG");
    if (getenv(ConfigFile) != NULL)
	strcpy(ConfigFile, getenv(ConfigFile));
    else {
    retry_config:	 
	 printf("\nEnter Configuration File Specification: ");
	 status=scanf("%s",ConfigFile);
	 fflush(stdin);
	 if (status==EOF) {
	    printf("\n\tCannot collect possible actual lookaside file\n");
	    goto done_config;
	 }
    }
    if ((gdcfile = fopen(ConfigFile, "r")) == (FILE *) NULL) {
	  printf("Cannot open file '%s'\n", ConfigFile);
	  goto retry_config;
     }
     while (fgets(ConfigFile, sizeof(ConfigFile), gdcfile)!= NULL) {
	  ZapCRLF(ConfigFile);
	  if (*ConfigFile == '#' || *ConfigFile == '\0') /** Ignore comments **/
	       continue;
	  cp = strchr(ConfigFile, ':');
	  if (cp == NULL) 
	       continue;
	  *cp = '\0';
	  token      = ConfigFile;
	  if (strncasecmp(token, "LookAside")==0) {
	    restofline = cp+1;
	    while (*restofline == ' ' || *restofline == '\t')
	       restofline++;
	    if ((cp = strchr(restofline, '!')) != NULL) {
		*cp = '\0';
		for(--cp;isspace(*cp)||!isprint(*cp);cp--)    
		    *cp = '\0';
		strcpy(lookaside, restofline);
	    }
	  }
     }
     fclose(gdcfile);
     if (strlen(lookaside)==0)
	goto retry_config;
done_config:
    acl_fab = cc$rms_fab;		/* Set up the FAB blocks */
    lks_fab = cc$rms_fab;
    lks_rab = cc$rms_rab;		/* Set up the RAB block */
    acl_xab = cc$rms_xabpro;		/* Set up the XAB block */
    acl_nam = cc$rms_nam;		/* Set up the NAM block */

    lks_fab.fab$l_ctx = acl_fab.fab$l_ctx = -1; /* User CTX=-1 Closed */
    acl_xab.xab$l_aclbuf = NULL;
    acl_xab.xab$w_aclsiz = 0;
    if (argc > 2)
	DEBUG = 1;
    command_file.dsc$b_class = input_file.dsc$b_class = DSC$K_CLASS_S;
    acl_fab.fab$l_fna = input_file.dsc$a_pointer = argv[1];
    acl_fab.fab$b_fns = input_file.dsc$w_length  = strlen(argv[1]);
    if (!(1&(status=sys$open(&acl_fab,0,0)))) {
	printf("\n\nCannot open file %s\n",argv[1]);
	exit(status);
    }
    sys$close(&acl_fab,0,0);
    command_file.dsc$a_pointer = (char *)malloc(5+strlen(argv[0]));
    strcpy(command_file.dsc$a_pointer, argv[0]);
    cp = strrchr(command_file.dsc$a_pointer,'.');
    strcpy(cp,".ini");
    command_file.dsc$w_length  = strlen(command_file.dsc$a_pointer);
    options = edt$m_nocreate;
    status = EDT$EDIT(&input_file, 0/*out_file*/, &command_file, 
				    0/*jou_file*/, &options, &fileio_bpv, 
					0/*workio*/, 0/*xlate*/);
    if (status != SS$_NORMAL)
	return(0);
    else {
	if (status == edt$_inpfilnex)
	    printf("\n\n%s does not exist\n",argv[1]);
	return(status);
    }
}
int
acl_fileio(int *code, int *stream, 
	    struct dsc$descriptor *record, struct dsc$descriptor *rhb)
{
/*  status codes */
globalvalue ss$_normal;
globalvalue edt$_nonstdfil;
globalvalue rms$_eof;
globalvalue rms$_normal;
/*  Possible values for *code	*/
globalvalue edt$k_open_input;
#define					EDT$K_OPEN_INPUT		0
globalvalue edt$k_open_output_seq;
#define					EDT$K_OPEN_OUTPUT_SEQ		1
globalvalue edt$k_open_output_noseq;
#define					EDT$K_OPEN_OUTPUT_NOSEQ		2
globalvalue edt$k_open_in_out;
#define					EDT$K_OPEN_IN_OUT		3
globalvalue edt$k_get;
#define					EDT$K_GET			4
globalvalue edt$k_put;
#define					EDT$K_PUT			5
globalvalue edt$k_close_del;
#define					EDT$K_CLOSE_DEL			6
globalvalue edt$k_close;
#define					EDT$K_CLOSE			7
/*  Possible values for *stream	*/
globalvalue edt$k_command_file;
#define					EDT$K_COMMAND_FILE		0
globalvalue edt$k_input_file;
#define					EDT$K_INPUT_FILE		1
globalvalue edt$k_include_file;
#define					EDT$K_INCLUDE_FILE		2
globalvalue edt$k_journal_file;
#define					EDT$K_JOURNAL_FILE		3
globalvalue edt$k_output_file;
#define					EDT$K_OUTPUT_FILE		4
globalvalue edt$k_write_file;
#define					EDT$K_WRITE_FILE		5

#define Codes 8
static	int	codes[Codes] = {edt$k_open_input, edt$k_open_output_seq, 
			edt$k_open_output_noseq, edt$k_open_in_out, 
			edt$k_get, edt$k_put, edt$k_close_del, edt$k_close};
#define Streams 6
static	int	streams[Streams] = {edt$k_command_file, edt$k_input_file,
			    edt$k_include_file, edt$k_journal_file,
			    edt$k_output_file, edt$k_write_file};
	int	CodeNum, StreamNum;
	int	status, g, linktuple;
	char    *g$, *acex$, *acey$;
struct	item	*itmlst;
struct	acedef  *ace$;
static	char    *INacl_ptr, *acex, *acey;
static	int	acl_len, acl_nlen, itmcnt=0;
static	char    sidename[256], buf[512];
	$DESCRIPTOR(dsc$buf,buf);

struct	Access
	{
	    struct Access *next$;
	    char    ace[];
	}   *aceA$, *aceL$;
    
    for (StreamNum=0; streams[StreamNum]!=*stream && StreamNum < Streams;
							StreamNum++);
    switch(StreamNum)
    {
    case EDT$K_INPUT_FILE:			    break;
    case EDT$K_OUTPUT_FILE:			    break;
    case EDT$K_JOURNAL_FILE:			    break;
    default:	return(EDT$FILEIO(code, stream, record, rhb));
    }
    for (CodeNum=0; codes[CodeNum]!=*code && CodeNum < Codes ; CodeNum++);
    switch(CodeNum)
    {
    case EDT$K_OPEN_INPUT:
    case EDT$K_OPEN_IN_OUT:
	    if (StreamNum==EDT$K_JOURNAL_FILE)
		return(ss$_normal);
	    acl_fab.fab$l_fna = record->dsc$a_pointer;	/* Set up the file name */
	    acl_fab.fab$b_fns = record->dsc$w_length;   /* for RMS to open */
	    acl_fab.fab$l_dna = rhb->dsc$a_pointer;
	    acl_fab.fab$b_dns = rhb->dsc$w_length;
	    acl_fab.fab$l_nam = &acl_nam;
	    acl_nam.nam$l_rsa = (char *)malloc(255);
	    acl_nam.nam$b_rss = 255;
	    status = sys$open(&acl_fab, 0, 0);	/* Open the file */
	    if (!(status&1)) {			/* Signal $OPEN errors */
		lib$signal(acl_fab.fab$l_sts, acl_fab.fab$l_stv);
		return(ss$_normal);
	    }
	    acl_fab.fab$l_ctx = 1;		/* User CTX=1 OPEN */
	    if (strlen(lookaside)) {
		*(acl_nam.nam$l_rsa+acl_nam.nam$b_rsl) = '\0';
		strcpy(sidename, acl_nam.nam$l_rsa);
		*(strchr(sidename, ']')) = '\0';
		strcat(sidename, ".");
		strcat(sidename, lookaside); 
		strcat(sidename, "]");
		strcat(sidename, strchr(acl_nam.nam$l_rsa, ']')+1);
		*(strchr(sidename, ';')) = '\0';
		lks_fab.fab$l_fna = sidename;
		lks_fab.fab$b_fns = strlen(sidename);   
		status = sys$open(&lks_fab, 0, 0);	/* Open the lookaside file */
		if (status&1) {
		    lks_fab.fab$l_ctx = 1;		/* User CTX=1 Open */
		    lks_rab.rab$l_fab = &lks_fab;
		    status = sys$connect(&lks_rab);
		    if (status&1)
			lks_rab.rab$b_rac = RAB$C_SEQ;
		}
	    }
	    return(ss$_normal);

    case EDT$K_OPEN_OUTPUT_SEQ:
    case EDT$K_OPEN_OUTPUT_NOSEQ:
		/*  Since we don't actually output to a file... */
	    return(ss$_normal);

    case EDT$K_GET:
	if (StreamNum==EDT$K_JOURNAL_FILE)
	    return(rms$_eof);
	switch(lks_fab.fab$l_ctx)
	{
	case 1:		/* Supply commentary header for the lookaside file */
	    lks_fab.fab$l_ctx = 2;	    /* User CTX=2 Continue lookaside */
	    sprintf(buf,"#from lookaside file %s #",sidename);
	    dsc$buf.dsc$w_length = strlen(buf);
	    dsc$buf.dsc$a_pointer = buf;
	    str$copy_dx(record, &dsc$buf);
	    return(ss$_normal);
	case 2:		/* Continue reading the lookaside file */
	    lks_rab.rab$l_ubf = buf;
	    lks_rab.rab$w_usz = sizeof(buf);
	    if ((status = sys$get(&lks_rab))==rms$_normal) {
		dsc$buf.dsc$w_length = lks_rab.rab$w_rsz;
		dsc$buf.dsc$a_pointer = lks_rab.rab$l_ubf;
		str$copy_dx(record, &dsc$buf);
		return(ss$_normal);
	    }
	    /*  Fall thru and do ACL setup *only* */
	case -1:	/* No Lookaside file present */
	    lks_fab.fab$l_ctx = 3;	    /* User CTX=3 EOF on lookaside */
	/*
	    Now that we have exhausted lookaside entries in any lookaside
	    file, we check for ACL entries in the original file.
	    Connect the XAB to the FAB block, allocate some ACL buffer
	    and perform an initial $DISPLAY.
	*/
	    acl_fab.fab$l_xab = &acl_xab;
	    INacl_ptr = acl_xab.xab$l_aclbuf = (char *)malloc(ACL_BUFFER_SIZE);
	    acl_xab.xab$w_aclsiz = ACL_BUFFER_SIZE;
	    acl_fab.fab$l_ctx = 2;
	    acl_xab.xab$l_aclctx = 0;
	    _zero(ACL_BUFFER_SIZE, acl_xab.xab$l_aclbuf);
	    status = sys$display(&acl_fab, 0, 0);
	    if (!(status&1)) {
		lks_fab.fab$l_ctx = 4;	    /* User CTX=4 "EOF" on ACLs   */
		return(rms$_eof);	    /* Report EOF to EDT	*/
	    }
	    sprintf(buf,"#Gopher ACE's from %s #",acl_nam.nam$l_rsa);
	    dsc$buf.dsc$w_length = strlen(buf);
	    dsc$buf.dsc$a_pointer = buf;
	    str$copy_dx(record, &dsc$buf);
	    return(ss$_normal);
	case 3:		/* Read in the ACL entries */
	/*  While we actually have an ACL and the ACL lookup was correct, we
	    run through the ACL list within the ACL buffer and format each ACE.
	    Since the buffer is relatively small, keep '$DISPLAY'ing the file 
	    until there is no more ACL entries to find.  The first $DISPLAY
	    was done when the file was $OPENed, the next $DISPLAY reads 
	    XAB$L_ACLCXT to get subsequent ACL entries for the buffer.
        */
	    while ((acl_xab.xab$w_acllen != 0)) {
		while((*INacl_ptr != 0) && ((INacl_ptr - acl_xab.xab$l_aclbuf) 
			    < ACL_BUFFER_SIZE)) {
			    /* The first byte is the size of the ACL */
		    ace$ = (struct acedef *)INacl_ptr;
		    INacl_ptr += *INacl_ptr;   /* Get the next ACL entry */
		    if ((ace$->ace$b_type==ACE$C_INFO) &&
			(strncmp((char *)&(ace$->ace$l_access),
				GOPHER_ACE,strlen(GOPHER_ACE))==0)) {
		    /* 
			This is one of our GOPHER Lookaside Application ACE's
		    */
			itmcnt++;
			if (acl_ptr==NULL) {
			    acex = acl_ptr = (char *)malloc(ACL_BUFFER_SIZE);
			    acl_len = ACL_BUFFER_SIZE;
			    _zero(ACL_BUFFER_SIZE, acl_ptr);
			}
			g = *((char *)ace$);
			if ((acex+g) > (acl_ptr+acl_len)) {
			    g = acex - acl_ptr;
			    acl_ptr = (char *)realloc(acl_ptr,
						acl_len+ACL_BUFFER_SIZE);
			    acl_len += ACL_BUFFER_SIZE;
			    acex = acl_ptr + g;
			    _zero(acl_len - (acex-acl_ptr), acex);
			    g = *((char *)ace$);
			}
			memcpy(acex, (char *)ace$, g);
			acex += g;
			g = ace$->ace$b_size - GOPHER_ACESZ;
			g$ = ((char *)&(ace$->ace$l_access)) 
					+ strlen(GOPHER_ACE);
			sprintf(buf,"%.*s",g, g$);
			dsc$buf.dsc$w_length = strlen(buf);
			dsc$buf.dsc$a_pointer = buf;
			str$copy_dx(record, &dsc$buf);
			return(ss$_normal);
		    }
		}
		_zero(ACL_BUFFER_SIZE, acl_xab.xab$l_aclbuf);
		status = sys$display(&acl_fab, 0, 0);/* Get the next ACL block */
		if (!(status&1)) break;
		if (!(acl_xab.xab$l_aclsts&1)) break;
	    }
	    acl_fab.fab$l_ctx = 4;	    /* User CTX=4 "EOF" on ACL's    */
	    return(rms$_eof);		    /* Report EOF to EDT	*/
	}
    case EDT$K_PUT:
	    if (StreamNum==EDT$K_JOURNAL_FILE)
		return(ss$_normal);
	    g = dsc$buf.dsc$w_length = sizeof(buf) - GOPHER_ACESZ - 1;
	    dsc$buf.dsc$a_pointer = buf + GOPHER_ACESZ;
	    str$copy_dx(&dsc$buf,record);
	    if (g > record->dsc$w_length)
		g = record->dsc$w_length;
	    buf[g + GOPHER_ACESZ] = '\0';
	    while (buf[strlen(buf)-1]==' ')
		buf[strlen(buf)-1] = '\0';
	    if (-1==(linktuple=LinkTuple(dsc$buf.dsc$a_pointer,NULL)))
		return(ss$_normal);	    /* Ignore non link-tuple elements */
	    /*
		This is a link tuple entry; now build an ACE from this line 
	    */
	    itmcnt++;
	    ace$ = (struct acedef *)buf;
	    ace$->ace$b_type = ACE$C_INFO;
	    ace$->ace$w_flags = 0;
	    memcpy((char *)&ace$->ace$l_access,GOPHER_ACE,strlen(GOPHER_ACE));
	    ace$->ace$b_size = strlen(dsc$buf.dsc$a_pointer) + GOPHER_ACESZ;
	    if (acl_new==NULL) {
		acey = acl_new = (char *)malloc(ACL_BUFFER_SIZE);
		acl_nlen = ACL_BUFFER_SIZE;
		_zero(ACL_BUFFER_SIZE, acl_new);
	    }
	    g = *((char *)ace$);
	    if ((acey+g+sizeof(char *)) > (acl_new+acl_nlen)) {
		g = acey - acl_new;
		acl_new = (char *)realloc(acl_new, acl_nlen+ACL_BUFFER_SIZE);
		acl_nlen += ACL_BUFFER_SIZE;
		acey = acl_new + g;
		_zero(acl_nlen - (acey-acl_new), acey);
		g = *((char *)ace$);
	    }
	    /* 
		Record this link tuple element from the edit buffer.
		If "ACCESS=", record into a linked list of multiple
		access elements; otherwise discard any earlier versions
		of the same element, and keep this one.
	    */
	    linktuple=LinkTuple(NULL,ace$);
	    if (linktuple==access_tuple) {
		aceA$ = (struct Access *)acey;
		aceA$->next$ = NULL;
		memcpy(aceA$->ace, (char *)ace$, g);
		if (g$ace[linktuple]==NULL)
		    g$ace[linktuple] = (char *)aceA$;
		else {
		    for (aceL$ = (struct Access *)g$ace[linktuple];
			    aceL$->next$; aceL$ = aceL$->next$);
			aceL$->next$ = aceA$;
		}
		acey += g + sizeof(struct Access *);
	    }
	    else {
		memcpy(acey, (char *)ace$, g);
		g$ace[linktuple] = acey;
		acey += g;
	    }
	    return(ss$_normal);
    case EDT$K_CLOSE_DEL:	    /* Close/delete -> Close	*/
    case EDT$K_CLOSE:
	    if (StreamNum == EDT$K_JOURNAL_FILE)
		return(ss$_normal);
	    if (StreamNum == EDT$K_OUTPUT_FILE) {
	    /*
		Build the SYS$CHANGE_ACL() item list.  First build Delete
		entry items for the old ACE's which are not still found
		in the edit buffer.  There will be at most itmcnt (+1) 
		item list entries.
	    */
		itmlst = (struct item *)malloc(sizeof(struct item)
						*(itmcnt + 1));
		itmcnt=0;
		if (acl_ptr==NULL)
		    goto    done_delete_list;
		for (acex$ = acl_ptr;
			((*acex$ != 0) && ((acex$ - acl_ptr) < acl_len));
			    acex$ += *acex$) {
		/*
		    This entry might exactly match an entry from the edit 
		    buffer; if so, just keep the original and ignore the new
		    (but unchanged) entry; otherwise add the original ACE to 
		    the delete list; always add original ACE's which aren't
		    in the edit buffer anymore to the delete list.
		*/
			linktuple = LinkTuple(NULL, (struct acedef *)acex$);
			if ((linktuple==-1) || (g$ace[linktuple]==NULL))
			    goto delete_item;
			/* Easiest to just delete prior "ACCESS=" entries */
			if (linktuple==access_tuple)
			    goto delete_item;
			if (memcmp(acex$, g$ace[linktuple], *acex$)==0) {
			    /* Keep the original, ignore unchanged new one  */
			    g$ace[linktuple] = NULL;
			}
			else
			{
			    /*  Build Delete item			    */
			delete_item:
			    itmlst[itmcnt].code = ACL$C_DELACLENT;
			    itmlst[itmcnt].bufadr = acex$;
			    itmlst[itmcnt].buflen = *acex$;
			    itmcnt++;
			}
		}
		done_delete_list:
	    /*
		Now for each non-null g$ace[] entry, build an Add item.
	    */
		for (linktuple=0; linktuple < GS_TUPLE_CNT; linktuple++) {
		    if ((acex$ = g$ace[linktuple])!=NULL) {
			if (linktuple==access_tuple) {
			    for (aceL$ = (struct Access *)g$ace[linktuple];
				    aceL$; aceL$ = aceL$->next$) {
				/* Build "ACCESS=" Add item */
				acex$ = aceL$->ace;
				itmlst[itmcnt].code = ACL$C_ADDACLENT;
				itmlst[itmcnt].bufadr = acex$;
				itmlst[itmcnt].buflen = *acex$;
				itmcnt++;
			    }
			}
			else {
			    /* Build non-"ACCESS=" Add item */
			    itmlst[itmcnt].code = ACL$C_ADDACLENT;
			    itmlst[itmcnt].bufadr = acex$;
			    itmlst[itmcnt].buflen = *acex$;
			    itmcnt++;
			}
		    }
		}
		g = itmlst[itmcnt].buflen = itmlst[itmcnt].code = 0;
		itmcnt = ACL$C_FILE;
		dsc$buf.dsc$a_pointer = acl_nam.nam$l_rsa;
		dsc$buf.dsc$w_length = strlen(acl_nam.nam$l_rsa);
		status = SYS$CHANGE_ACL(0, &itmcnt, &dsc$buf, itmlst, 0, 0, &g);
		if (acl_ptr != NULL) /* Free bufs when closing output file */
		    free(acl_ptr);
		if (acl_new != NULL)
		    free(acl_new);
	    }
	    if (lks_fab.fab$l_ctx != -1) {
		sys$close(&lks_fab, 0, 0);
		lks_fab.fab$l_ctx = -1;	    /* User CTX=-1 Closed */
	    }
	    if (acl_xab.xab$l_aclbuf != NULL) {
		free(acl_xab.xab$l_aclbuf);
		acl_xab.xab$l_aclbuf = NULL;
		acl_xab.xab$w_aclsiz = 0;
	    }
	    if (acl_fab.fab$l_ctx != -1) {
		sys$close(&acl_fab, 0, 0);
		acl_fab.fab$l_ctx = -1;	    /* User CTX=-1 Closed */
	    }
	    return(ss$_normal);
    default:	return(EDT$FILEIO(code, stream, record, rhb));
    }
}

int
LinkTuple(char *t, struct acedef *ace)
{
    int	lt;
static
    char *keyword[GS_TUPLE_CNT] = { GS_NAME, GS_TYPE, GS_PATH, GS_HOST, 
				    GS_PORT, GS_NUMB, GS_HDDN, GS_ACCS,
				    GS_HEAD, GS_FOOT, GS_RHEAD, GS_RFOOT };

    if (ace != NULL)
	t = (char *)&ace->ace$l_access + strlen(GOPHER_ACE);
    for (lt=0; lt < GS_TUPLE_CNT; lt++)
	if (strncasecmp(t, keyword[lt], strlen(keyword[lt]))==0)
	    return(lt);
    return(-1);
}
